import logging

from flask_appbuilder.security.manager import AUTH_OID
from flask_appbuilder.security.manager import AUTH_OAUTH
from flask_appbuilder.security.sqla.manager import SecurityManager
from flask_oidc import OpenIDConnect

from .oidc_views import DynamicRoleAuthOIDCView
from .oauth_views import DynamicRoleAuthOAuthView


log = logging.getLogger(__name__)


class DynamicRoleSecurityManagerMixin:
    def __init__(self, appbuilder):
        super().__init__(appbuilder)
        if self.auth_type == AUTH_OID:
            self.oid = OpenIDConnect(self.appbuilder.get_app)
            self.authoidview = DynamicRoleAuthOIDCView
        elif self.auth_type == AUTH_OAUTH:
            self.authoauthview = DynamicRoleAuthOAuthView


try:
    from superset.security import SupersetSecurityManager
    class SupersetOIDCSecurityManager(DynamicRoleSecurityManagerMixin,
                                      SupersetSecurityManager):
        pass

    class SupersetOAuthSecurityManager(DynamicRoleSecurityManagerMixin,
                                       SupersetSecurityManager):
        def __init__(self, appbuilder):
            super().__init__(appbuilder)
            if self.auth_type == AUTH_OAUTH:
                self.oauth_user_info = self.get_oauth_user_info

        def get_oauth_user_info(self, provider, resp):
            if provider != 'azure':
                return super().get_oauth_user_info(provider, resp)

            log.error("message from fab_auth_dynamic_roles")
            log.error("Azure response received : {0}".format(resp))
            id_token = resp["id_token"]
            log.error(str(id_token))
            me = self._azure_jwt_token_parse(id_token)
            log.error("Parse JWT token : {0}".format(me))
            # must set username from email and email from email for migration from OIDC to OAuth
            return {
                "name": me.get("name", ""),
                "email": me["email"],
                "first_name": me.get("given_name", ""),
                "last_name": me.get("family_name", ""),
                "id": me["oid"],
                "username": me["email"],
                "roles": me.get("roles", []),
            }

except ImportError:
    log.error("from superset.security import SupersetSecurityManager failed")


try:
    from airflow.www.security import AirflowSecurityManager
    class AirflowOIDCSecurityManager(DynamicRoleSecurityManagerMixin,
                                     AirflowSecurityManager):
        pass
    class AirflowOAuthSecurityManager(DynamicRoleSecurityManagerMixin,
                                      AirflowSecurityManager):
        pass

except ImportError:
    log.error("from airflow.www_rbac.security import AirflowSecurityManager failed")


try:
    from astronomer.flask_appbuilder.security import AirflowAstroSecurityManager
    class AirflowAstroOIDCSecurityManager(DynamicRoleSecurityManagerMixin,
                                          AirflowAstroSecurityManager):
        pass
    class AirflowAstroOAuthSecurityManager(DynamicRoleSecurityManagerMixin,
                                          AirflowAstroSecurityManager):
        pass

except ImportError:
    log.error("from astronomer.flask_appbuilder.security import AirflowAstroSecurityManager failed")



import logging
import os
import urllib

from flask import redirect, request
from flask_admin import expose
from flask_appbuilder.security.views import AuthOIDView
from flask_login import login_user


USERNAME_OIDC_FIELD = os.getenv(
    'USERNAME_OIDC_FIELD',
    default='preferred_username',
)

FIRST_NAME_OIDC_FIELD = os.getenv(
    'FIRST_NAME_OIDC_FIELD',
    default='given_name',
)

LAST_NAME_OIDC_FIELD = os.getenv(
    'LAST_NAME_OIDC_FIELD',
    default='family_name',
)

EMAIL_OIDC_FIELD = os.getenv(
    'EMAIL_OIDC_FIELD',
    default='email',
)

CLIENT_ROLE_OIDC_FIELD = os.getenv(
    'CLIENT_ROLE_OIDC_FIELD',
    default='roles',
)

RESOURCE_ACCESS_APP_OIDC_FIELD = os.getenv(
    'RESOURCE_ACCESS_APP_OIDC_FIELD',
    default='superset',
)

OIDC_LOGOUT_URI = 'OIDC_LOGOUT_URI'


log = logging.getLogger(__name__)


class DynamicRoleAuthOIDCView(AuthOIDView):

    @expose('/login/', methods=['GET', 'POST'])
    def login(self, flag=True):

        sm = self.appbuilder.sm
        oidc = sm.oid

        @self.appbuilder.sm.oid.require_login
        def handle_login():
            user = sm.auth_user_oid(oidc.user_getfield(EMAIL_OIDC_FIELD))
            info = oidc.user_getinfo([
                USERNAME_OIDC_FIELD,
                FIRST_NAME_OIDC_FIELD,
                LAST_NAME_OIDC_FIELD,
                EMAIL_OIDC_FIELD,
                CLIENT_ROLE_OIDC_FIELD,
            ])

            if user is None:
                user = sm.add_user(
                    username=info.get(USERNAME_OIDC_FIELD),
                    first_name=info.get(FIRST_NAME_OIDC_FIELD),
                    last_name=info.get(LAST_NAME_OIDC_FIELD),
                    email=info.get(EMAIL_OIDC_FIELD),
                    role=sm.find_role(sm.auth_user_registration_role)
                )
                log.info(f"user added: {info.get(EMAIL_OIDC_FIELD)}")

            # sync roles from OIDC to flask
            user.roles.clear()
            sm.update_user(user)

            if info.get(CLIENT_ROLE_OIDC_FIELD) is None:
                log.error(f'user {info.get(EMAIL_OIDC_FIELD)} does not have ROLE')
                log.error(f'user info: {info}')
            else:
                for role_name in info.get(CLIENT_ROLE_OIDC_FIELD):
                    role = sm.find_role(role_name)
                    if role is not None:
                        user.roles.append(role)
                        log.info(f"assign role: {role_name}, find_role: {role} to user: {userinfo.get(EMAIL_OAUTH_FIELD)}")
                    else:
                        log.error(f"role: {role_name} doesn't exist")
                sm.update_user(user)

            login_user(user, remember=False)
            return redirect(self.appbuilder.get_url_for_index)

        return handle_login()

    @expose('/logout/', methods=['GET', 'POST'])
    def logout(self):
        oidc = self.appbuilder.sm.oid
        oidc.logout()
        super(DynamicRoleAuthOIDCView, self).logout()

        request_root = request.url_root.strip('/')
        redirect_url = f'{request_root}{self.appbuilder.get_url_for_login}'

        issuer = oidc.client_secrets.get('issuer')
        logout_uri = f'{issuer}/protocol/openid-connect/logout?redirect_uri='
        if OIDC_LOGOUT_URI in self.appbuilder.app.config:
            logout_uri = self.appbuilder.app.config[OIDC_LOGOUT_URI]

        return redirect(f'{logout_uri}{urllib.parse.quote(redirect_url)}')



import logging
import os
import re
import urllib

from flask import flash, g, redirect, request, session, url_for
from flask_appbuilder._compat import as_unicode
from flask_appbuilder.baseviews import expose
from flask_appbuilder.security.views import AuthOAuthView
from flask_login import login_user
import jwt


EMAIL_OAUTH_FIELD = os.getenv(
    'EMAIL_OAUTH_FIELD',
    default='email',
)

CLIENT_ROLE_OAUTH_FIELD = os.getenv(
    'CLIENT_ROLE_OAUTH_FIELD',
    default='roles',
)


log = logging.getLogger(__name__)


# TODO
# - AzureAD의 경우 Group 정보 얻어와서 인증을 좀 더 부드럽게

class DynamicRoleAuthOAuthView(AuthOAuthView):

    @expose("/login/")
    @expose("/login/<provider>")
    @expose("/login/<provider>/<register>")
    def login(self, provider=None, register=None):
        log.debug("Provider: {0}".format(provider))
        if g.user is not None and g.user.is_authenticated:
            log.debug("Already authenticated {0}".format(g.user))
            return redirect(self.appbuilder.get_url_for_index)

        if provider is None:
            if len(self.appbuilder.sm.oauth_providers) > 1:
                return self.render_template(
                    self.login_template,
                    providers=self.appbuilder.sm.oauth_providers,
                    title=self.title,
                    appbuilder=self.appbuilder,
                )
            else:
                provider = self.appbuilder.sm.oauth_providers[0]["name"]

        log.debug("Going to call authorize for: {0}".format(provider))
        state = jwt.encode(
            request.args.to_dict(flat=False),
            self.appbuilder.app.config["SECRET_KEY"],
            algorithm="HS256",
        )
        try:
            if register:
                log.debug("Login to Register")
                session["register"] = True
            if provider == "twitter":
                return self.appbuilder.sm.oauth_remotes[
                    provider
                ].authorize_redirect(
                    redirect_uri=url_for(
                        ".oauth_authorized",
                        provider=provider,
                        _external=True,
                        state=state,
                    )
                )
            else:
                return self.appbuilder.sm.oauth_remotes[
                    provider
                ].authorize_redirect(
                    redirect_uri=url_for(
                        ".oauth_authorized", provider=provider, _external=True
                    ),
                    state=state.decode("ascii")
                    if isinstance(state, bytes)
                    else state,
                )
        except Exception as e:
            log.error("Error on OAuth authorize: {0}".format(e))
            flash(as_unicode(self.invalid_login_message), "warning")
            return redirect(self.appbuilder.get_url_for_index)

    @expose("/oauth-authorized/<provider>")
    def oauth_authorized(self, provider):
        log.debug("Authorized init")
        resp = self.appbuilder.sm.oauth_remotes[provider].authorize_access_token()
        if resp is None:
            flash(u"You denied the request to sign in.", "warning")
            return redirect(self.appbuilder.get_url_for_login)

        log.debug("OAUTH Authorized resp: {0}".format(resp))
        # Retrieves specific user info from the provider
        try:
            self.appbuilder.sm.set_oauth_session(provider, resp)
            userinfo = self.appbuilder.sm.oauth_user_info(provider, resp)
        except Exception as e:
            log.error("Error returning OAuth user info: {0}".format(e))
            user = None
        else:
            log.debug("User info retrieved from {0}: {1}".format(provider, userinfo))
            # User email is not whitelisted
            if provider in self.appbuilder.sm.oauth_whitelists:
                whitelist = self.appbuilder.sm.oauth_whitelists[provider]
                allow = False
                for e in whitelist:
                    if re.search(e, userinfo["email"]):
                        allow = True
                        break
                if not allow:
                    flash(u"You are not authorized.", "warning")
                    return redirect(self.appbuilder.get_url_for_login)
            else:
                log.debug("No whitelist for OAuth provider")
            user = self.appbuilder.sm.auth_user_oauth(userinfo)

        if user is None:
            flash(as_unicode(self.invalid_login_message), "warning")
            return redirect(self.appbuilder.get_url_for_login)
        else:
            # sync roles from OAuth to flask
            sm = self.appbuilder.sm

            user.roles.clear()
            sm.update_user(user)

            if userinfo.get(CLIENT_ROLE_OAUTH_FIELD) is None:
                log.error(f'userinfo does not have {CLIENT_ROLE_OAUTH_FIELD} field')
                log.error(f'user info: {userinfo}')
            else:
                for role_name in userinfo.get(CLIENT_ROLE_OAUTH_FIELD):
                    role = sm.find_role(role_name)
                    if role is not None:
                        user.roles.append(role)
                        log.info(f"assign role: {role_name}, find_role: {role} to user: {userinfo.get(EMAIL_OAUTH_FIELD)}")
                    else:
                        log.error(f"role: {role_name} doesn't exist")
                sm.update_user(user)

            login_user(user)
            try:
                state = jwt.decode(
                    request.args["state"],
                    self.appbuilder.app.config["SECRET_KEY"],
                    algorithms=["HS256"],
                )
            except jwt.InvalidTokenError:
                raise Exception("State signature is not valid!")

            try:
                next_url = state["next"][0] or self.appbuilder.get_url_for_index
            except (KeyError, IndexError):
                next_url = self.appbuilder.get_url_for_index

            return redirect(next_url)



# -*- coding: utf-8 -*-
"""fab-auth-dynamic-roles

#### A plugin for authentication to sync roles dynamically

#### Supports

##### Auth Modes

- AUTH_OID
- AUTH_OAUTH

##### Applications

- Superset

"""


_major_v = '0'
_minor_v = '1.3'


from os import path
import pathlib
import sys

import pkg_resources
from setuptools import setup, find_packages


if sys.version_info[:3] < (3, 6):
    raise RuntimeError("Python version 3.6 or later required.")


with pathlib.Path('requirements.txt').open() as rin:
    requirements = [str(req) for req in pkg_resources.parse_requirements(rin)]


setup(
    name='fab-auth-dynamic-roles',
    version=f'{_major_v}.{_minor_v}',
    description='Flask AppBuilder Authentication plugin',
    url='https://github.com/hyunjong-lee/fab-auth-dynamic-roles',
    author='Hyunjong Lee',
    author_email='hyunjong.lee.s@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=[
        'fab_auth_dynamic_roles',
    ],
    python_requires='>=3.6',
    install_requires=requirements,
    test_suite='nose.collector',
    tests_require=['nose'],
    entry_points={},
    project_urls={
        'Source': 'https://github.com/hyunjong-lee/fab-auth-dynamic-roles',
    },
    download_url=f'https://github.com/hyunjong-lee/fab-auth-dynamic-roles/archive/v{_major_v}.{_minor_v}.tar.gz',
    keywords=['dynamic roles', 'flask appbuilder', 'keycloak', 'AzureAD', 'superset', 'AUTH_OID', 'AUTH_OAUTH', "OAUTH", "OIDC"],
    license='LGPLv2+',
)
